{ kompresja plikow adaptacyjna metoda Huffmana
  autor implementacji: Michal Czardybon }

program Huffman;

const
{ domyslne wartosci parametrow kompresji }
 _NewTreeFreq=256;
 _FreqBufSize=2048;
 _InitialFreq=1;
{ maksymalna wartosc FreqBufSize }
 MaxFreqBufSize=8192;
{ wezel drzewa kodow Huffmana }
type TNode=
 record
  Freq,Left,Right,Parent,Route,Code:integer;
  Bit:byte;
 end;

{ zawiera kody Huffmana }
type THufTree=
 record
  Node:array[0..511] of TNode;
  Root:integer; {korzen}
 end;

{ wskaznik potrzebny do przekazywania drzewa do funkcji }
type PTHufTree=^THufTree;

var

{ parametry kompresji }
NewTreeFreq,         { jak czesto jest tworzone nowe drzewo }
FreqBufSize,         { rozmiar bufora dla czestosci bajtow }
InitialFreq:integer; { poczatkowa czestosc bajtow }

{  dane dotyczace czestosci wystepowania kodow }
ItemFreq:array [0..255] of integer;
FreqBuf:array [0..MaxFreqBufSize-1] of byte;
FreqBufPos:integer;

{ input / output }
fin,fout:file;
b,bp:byte;

{ starta }
{$I hufheap.pas}

{ tworzenie drzewa kodow }
procedure CreateHufTree(Huf:PTHufTree);
var
 i,a,b,nNodes:integer;
 HeapItem:array[0..256] of integer;
 HeapSize:integer;
begin
 { inicjowanie zmiennych }
 for i:=0 to 255 do
  begin
   Huf^.Node[i].Route:=-1;
   Huf^.Node[i].Code:=i;
   Huf^.Node[i].Freq:=ItemFreq[i];
   HeapItem[i+1]:=i;
  end;
 nNodes:=256;
 Huf^.Node[511].Freq:=-1;
 HeapItem[0]:=511;
 HeapSize:=256;
 { utworzenie sterty }
 HeapConstruct(Huf, HeapItem, HeapSize);
 while HeapSize>1 do
  begin
   { wez dwa wezly o najmniejszej czestosci }
   a:=HeapGetRoot(Huf, HeapItem, HeapSize);
   b:=HeapGetRoot(Huf, HeapItem, HeapSize);
   { utworz nowy wezel - rodzica tych dwoch }
   Huf^.Node[nNodes].Left:=a;
   Huf^.Node[nNodes].Right:=b;
   Huf^.Node[nNodes].Code:=-1;
   Huf^.Node[nNodes].Freq := Huf^.Node[a].Freq + Huf^.Node[b].Freq;
   Huf^.Node[a].Parent:=nNodes;
   Huf^.Node[b].Parent:=nNodes;
   Huf^.Node[a].Bit:=0;
   Huf^.Node[b].Bit:=1;
   { dodaj nowy wezel na sterte }
   HeapInsert(Huf, HeapItem, HeapSize, nNodes);
   inc(nNodes);
  end;
 { pobierz wierzcholek drzewa }
 Huf^.Root:=HeapGetRoot(Huf, HeapItem, HeapSize);
 { zaznacz, ze nie ma rodzica }
 Huf^.Node[Huf^.Root].Parent:=-1;
end;

{ aktualizacja danych dotyczacych czestosci bajtow }
procedure UpdateFreqData(c:byte);
begin
 dec(ItemFreq[ FreqBuf[FreqBufPos] ]);
 FreqBuf[FreqBufPos]:=c;
 inc(ItemFreq[ FreqBuf[FreqBufPos] ]);
 inc(FreqBufPos);
 if FreqBufPos=FreqBufSize then FreqBufPos:=0;
end;

{ skierowanie zakodowanego bajtu na wyjscie }
procedure OutputHufCode(Huf:PTHufTree; c:byte);
var m,n:integer;
begin
 { przejdz po drzewie od korzenia do wezla terminalnego, zaznaczajac droge }
 n:=c;
 while n<>Huf^.Root do
  begin
   m:=n;
   n:=Huf^.Node[n].Parent;
   Huf^.Node[n].Route:=m;
  end;
 { zapisuj po 1 bicie }
 n:=Huf^.Node[n].Route;
 while n<>-1 do
  begin
   b:=b + (Huf^.Node[n].Bit shl bp);
   inc(bp);
   if bp=8 then
    begin
     BlockWrite(fout,b,1);
     bp:=0;
     b:=0;
    end;
   n:=Huf^.Node[n].Route;
  end;
end;

{ kompresja }
procedure Compress;
var
 HufTree:PTHufTree;
 i,a:integer;
 c:byte;
 fs:longint;
begin
 { parametry kompresji = parametry domyslne }
 NewTreeFreq:=_NewTreeFreq;
 FreqBufSize:=_FreqBufSize;
 InitialFreq:=_InitialFreq;

 New(HufTree);
 { zapisanie dlugosci pliku wejsciowego }
 fs:=FileSize(fin);
 BlockWrite(fout,fs,4);
 { zapisanie parametrow kompresji }
 BlockWrite(fout,NewTreeFreq,2);
 BlockWrite(fout,FreqBufSize,2);
 BlockWrite(fout,InitialFreq,2);

 { inicjacja danych }
 for i:=0 to 255 do ItemFreq[i]:=InitialFreq;
 { przepisanie pierwszej serii bajtow - dane niezakodowane potrzebne do
   stworzenia pierwszego drzewa kodow }
 i:=0;
 while (i<FreqBufSize) and (not Eof(fin)) do
  begin
   BlockRead(fin,c,1);
   FreqBuf[i]:=c;
   inc(ItemFreq[c]);
   BlockWrite(fout,c,1);
   inc(i);
  end;
 FreqBufPos:=0;
 HufTree^.Root:=-1;
 a:=NewTreeFreq;
 b:=0; bp:=0;
 { wlasciwa kompresja }
 while not Eof(fin) do
  begin
   BlockRead(fin,c,1);
   if a=NewTreeFreq then { czas na stworzenie nowego drzewa }
    begin
     CreateHufTree(HufTree);
     a:=0;
    end;
   OutputHufCode(HufTree, c);
   UpdateFreqData(c);
   inc(a); { zwiekszenie "wieku" drzewa }
  end;
 { zapisanie ostatniego bajtu }
 if bp<>0 then BlockWrite(fout,b,1);
 Dispose(HufTree);
end;

{ odkodowanie jednego bajtu }
function DecodeNext(Huf:PTHufTree):integer;
var
 n:integer;
begin
 { przechodzenie drzewa od korzenia }
 n:=Huf^.Root;
 repeat
  if bp=8 then
   begin
    BlockRead(fin,b,1);
    bp:=0;
   end;
  if ((b shr bp) and 1) = 1 then
   n:=Huf^.Node[n].Right
  else
   n:=Huf^.Node[n].Left;
  inc(bp);
 until Huf^.Node[n].Code<>-1;
 DecodeNext:=Huf^.Node[n].Code;
end;

{ dekompresja }
procedure Decompress;
var
 HufTree:PTHufTree;
 fs,j:longint;
 i,a:integer;
 c:byte;
begin
 { odczytaj dlugosc pliku sprzed kompresji }
 BlockRead(fin,fs,4);
 { odczytaj parametry uzyte podczas kompresji }
 BlockRead(fin,NewTreeFreq,2);
 BlockRead(fin,FreqBufSize,2);
 BlockRead(fin,InitialFreq,2);

 if FreqBufSize<=MaxFreqBufSize then
  begin
   New(HufTree);
   { inicjacja zmiennych }
   for i:=0 to 255 do ItemFreq[i]:=InitialFreq;
   { przepisanie pierwszej serii danych - dane niezakodowane }
   i:=0;
   while (i<FreqBufSize) and (not Eof(fin)) do
    begin
     BlockRead(fin,c,1);
     FreqBuf[i]:=c;
     inc(ItemFreq[c]);
     BlockWrite(fout,c,1);
     inc(i);
    end;
   FreqBufPos:=0;
   HufTree^.Root:=-1;
   a:=NewTreeFreq;
   b:=0; bp:=8;
   { wlasciwa dekomresja }
   for j:=i to fs-1 do
    begin
     if a=NewTreeFreq then
      begin
       CreateHufTree(HufTree);
       a:=0;
      end;
     c:=DecodeNext(HufTree);
     BlockWrite(fout,c,1);
     UpdateFreqData(c);
     inc(a);
    end;
   Dispose(HufTree);
  end
 else
  writeln('Dekompresja anulowana.');
end;

begin
 if ParamCount<>3 then
  begin
   WriteLn('Kompresja plikow metoda Huffmana');
   WriteLn('autor: Michal Czardybon');
   WriteLn('skladnia:');
   WriteLn('huf.exe k|d plik1 plik2');
   WriteLn('k - kompresja, d - dekompresja');
   WriteLn('plik1 - plik wejsciowy, plik2 - plik wyjsciowy');
  end
 else
  begin
   Assign(fin,ParamStr(2)); Reset(fin,1);
   Assign(fout,ParamStr(3)); Rewrite(fout,1);
   if ParamStr(1)='k' then
    Compress
   else if ParamStr(1)='d' then
    Decompress;
   Close(fin);
   Close(fout);
  end;
end.